﻿"use strict";

function Checkbox(element, config) {
	if (typeof element === 'string')
		element = document.getElementById(element);
	
	var initial = config ? !!config.value : false;
	var current = initial;
	element.checked = current;
	
	var onChange = config && config.onChange ? config.onChange : null;
	
	element.oninput = function() {
		if (current === element.checked)
			return;
			
		current = element.checked;
		if (onChange)
			onChange();
	};
		
	return {
		getValue: function() { return current; },
		
		setValue: function(value, raise) {
			if (current === value)
				return;
		
			current = value;
			element.checked = current;
			if (onChange && raise)
				onChange();
		},
		
		reset: function(raise) {
			this.setValue(initial, raise);
		}
	};
}

function RadioGroup(elements, config) {
	for (var i = 0; i < elements.length; i++)
		if (typeof elements[i] === 'string')
			elements[i] = document.getElementById(elements[i]);
	
	var initial = config && 'value' in config ? config.value : 0;
	var current = initial;
	function updateChecked() {
		for (var i = 0; i < elements.length; i++)
			elements[i].checked = (i === current);
	}
	
	updateChecked();
	
	var onChange = config && config.onChange ? config.onChange : null;
		
	for (var i = 0; i < elements.length; i++)	
		elements[i].oninput = function() {
			if (!this.checked)
				return;
		
			var index = elements.indexOf(this);
			if (current === index)
				return;
				
			current = index;
			updateChecked();
			
			if (onChange)
				onChange();
		};
		
	return {
		getValue: function() { return current; },
		
		setValue: function(value, raise) {
			if (current === value)
				return;
		
			current = value;
			updateChecked();
			
			if (onChange && raise)
				onChange();
		},
		
		reset: function(raise) {
			this.setValue(initial, raise);
		}
	};
}

function Select(element, items, config) {
	if (typeof element === 'string')
		element = document.getElementById(element);

	var onChange = config && config.onChange ? config.onChange : null;

	for (var i = 0; i < items.length; i++) {
		var option = document.createElement('option');
		option.text = items[i].toString();
		element.add(option);
	}
	
	var initial = config && 'selected' in config ? config.selected : 0;
	var current = initial;
	element.selectedIndex = current;
	
	element.oninput = function() {
		if (current === element.selectedIndex)
			return;
	
		current = element.selectedIndex;
		if (onChange)
			onChange();
	};
	
	return {
		getSelectedIndex: function() { return current; },
		getSelected: function() { return items[current]; },
		
		setSelectedIndex: function(value, raise) {
			if (value === current)
				return;
				
			current = value;
			element.selectedIndex = current;
				
			if (onChange && raise)
				onChange();
		},
		
		unselect: function(raise) {
			this.setSelectedIndex(-1, raise);
		},
		
		reset: function(raise) {
			this.setSelectedIndex(initial, raise);
		},
		
		enable: function(state) {
			element.disabled = !state;
		}
	};
}

function IntSlider(element, min, max, config) {
	if (typeof element === 'string')
		element = document.getElementById(element);

	var onChange = config && config.onChange ? config.onChange : null;
	
	var initial = config && config.value ? config.value : min;
	var current = initial;
	
	element.min = min;
	element.max = max;
	element.value = current;
	
	element.oninput = function() {
		var value = parseInt(element.value);
		if (current === value)
			return;
	
		current = value;
		
		display();
		if (onChange)
			onChange();
	};
	
	var display = null;
	if (config && config.display) {
		var de = config.display;
		if (typeof de === 'string')
			de = document.getElementById(de);
			
		display = function() {
			de.innerHTML = current.toString();
		};
		
		display();
	}
	else
		display = function() {  };
	
	return {
		getValue: function() { return current; },
		
		setValue: function(value, raise) {
			if (current === value)
				return;
		
			current = value;
			element.value = current;
			display();
			
			if (onChange && raise)
				onChange();
		},
		
		reset: function(raise) {
			this.setValue(initial, raise);
		}
	};
}

function Slider(element, min, max, config) {
	if (typeof element === 'string')
		element = document.getElementById(element);
	
	var initial = config && 'value' in config ? config.value : min;
	initial = Math.constrain(initial, min, max);
	var current = initial;
	
	var onChange = config && config.onChange ? config.onChange : null;
	var format = config && config.format ? config.format : PreciseFormat(4);
	
	var	steps = 1000;
	if (config && config.steps)
		steps = config.steps;
	else if (config && config.increment)
		steps = Math.round((max - min) / config.increment);
	
	element.min = 0;
	element.max = steps;
		
	var setScaled;
	var getScaled;
	if (config && config.log) {
		var ratio = max / min;
	
		setScaled = function(value) {
			element.value = steps * Math.log(value / min) / Math.log(ratio);
		};
		
		getScaled = function() {
			var value = parseInt(element.value);
			return min * Math.pow(ratio, value / steps);
		};
	} else {
		var diff = max - min;
	
		setScaled = function(value) {
			element.value = steps * (value - min) / diff;
		};
		
		getScaled = function() {
			var value = parseInt(element.value);
			return min + diff * value / steps;
		};
	}
		
	setScaled(current);
	element.oninput = function() {
		var value = getScaled();
		if (current === value)
			return;
	
		current = value;
		display();
		
		if (onChange)
			onChange();
	};
	
	var display;
	if (config && config.display) {
		var de = config.display;
		if (typeof de === 'string')
			de = document.getElementById(de);
			
		display = function() {
			de.innerHTML = format(current);
		};
		
		display();
	}
	else
		display = function() {  };
	
	return {
		getValue: function() { return current; },
		
		setValue: function(value, raise) {
			if (value > max)
				value = max;
			else if (value < min)
				value = min;
		
			if (current === value)
				return;
		
			current = value;
			setScaled(current);
			display();
			
			if (onChange && raise)
				onChange();
		},
		
		mult: function(amount, raise) {
			this.setValue(amount * current, raise);
		},
		
		shift: function(amount, raise) {
			this.setValue(current + amount, raise);
		},
		
		reset: function(raise) {
			this.setValue(initial, raise);
		},
		
		setInitial: function(value) {
			if (value < min)
				initial = min;
			else if (value > max)
				initial = max;
			else
				initial = value;
		},
		
		enable: function(state) {
			element.disabled = !state;
		}
	};
}

