﻿"use strict";

var Fourier = (function(){
	function reduce(x) {
		var count = Math.floor(x / Math.DPI);
		return x - count * Math.DPI;
	};

	function F(data) {
		this.eval = data.eval;
	
		this.getComp = function(n) {
			var amp = data.amp(n);
			if (!amp)
				return Math.zero;
			
			var freq = data.freq(n);
			if (!freq)
				return Math.constant(amp);
			
			var phase = data.phase(n);
			
			return function(wt) {
				return amp * Math.cos(freq * wt + phase);
			};
		};
		
		this.getPartial = function(order) {
			var comp = [];
			for (var n = 0; n <= order; n++)
				comp.push( this.getComp(n) );
		
			return function(wt) {
				var result = 0;
				for (var n = 0; n <= order; n++)
					result += comp[n](wt);
				return result;
			};
		};
		
		this.getAmp = data.amp;
		this.getPhase = data.phase;
		this.getFreq = data.freq;
		
		this.getMin = function() { return 'min' in data ? data.min : -1; };
		this.getMax = function() { return 'max' in data ? data.max : 1; };
	}

	F.Square = function() {
		return new F({
			eval: function(wt) { return reduce(wt) < Math.PI ? 1 : -1; },
			freq: function(n) { return n ? 2 * n - 1 : 0; },
			amp: function(n) { return n ? 4 / (Math.PI * (2 * n - 1)) : 0; },
			phase: function(n) { return n ? -Math.HPI : 0; }
		});
	};

	F.Rectif = function() {
		var AVERAGE = 2 / Math.PI;
	
		return new F({
			eval: function(wt) { return Math.abs(Math.sin(wt)); },
			freq: function(n) { return 2 * n; },
			amp: function(n) { return n ? 2 * AVERAGE / (4 * n * n - 1) : AVERAGE; },
			phase: function(n) { return n ? Math.PI : 0; },
			min: 0
		});
	};

	F.Sawtooth = function() {
		return new F({
			eval: function(wt) { return reduce(wt) / Math.PI - 1; },
			freq: function(n) { return n; },
			amp: function(n) { return n ? 2 / (Math.PI * n) : 0; },
			phase: function(n) { return n ? Math.HPI : 0; }
		});
	};

	F.Triangle = function() {
		return new F({
			eval: function(wt) {
				var u = reduce(wt) / Math.PI;
				return u < 1 ? 1 - 2 * u : 2 * u - 3;
			},
			freq: function(n) { return n ? 2 * n - 1 : 0; },
			amp: function(n) { return n ? 8 / Math.pow(Math.PI * (2 * n - 1), 2) : 0; },
			phase: function(n) { return 0; }
		});
	};

	F.Pulse = function(duty) {
		var THRESHOLD = Math.DPI * duty;
		var SHIFT = Math.PI * duty;
	
		return new F({
			eval: function(wt) { return reduce(wt + SHIFT) < THRESHOLD ? 1 : 0; },
			freq: function(n) { return n; },
			amp: function(n) { return n ? 2 * Math.abs(Math.sin(n * SHIFT)) / (Math.PI * n) : duty; },
			phase: function(n) { return Math.sin(n * SHIFT) < 0 ? Math.PI : 0; },
			min: 0
		});
	};

	return F;
})();