﻿"use strict";

var TRACE_POINTS = 1000;

var PLOTLY = (function(){
	var FONT = { size: 20, color: 'black' };
	var MARGIN_HORZ = 60;
	var MARGIN_HORZ_EMPTY = 7;
	var MARGIN_VERT = 90;
	var MARGIN_VERT_EMPTY = 16;
	var THIN_LINE = 1;
	var NORMAL_LINE = 1.5;
	var THICK_LINE = 2.5;
	var AXIS_LINE = 1.25;

	var XRANGE0 = 'xaxis.range[0]';
	var XRANGE1 = 'xaxis.range[1]';
	var YRANGE0 = 'yaxis.range[0]';
	var YRANGE1 = 'yaxis.range[1]';

	function RangeManager(left, right, bottom, top) {
		var x0 = left;
		var x1 = right;
		var y0 = bottom;
		var y1 = top;
		
		function relayout(data) {
			if (XRANGE0 in data)
				x0 = data[XRANGE0];
				
			if (XRANGE1 in data)
				x1 = data[XRANGE1];
				
			if (YRANGE0 in data)
				y0 = data[YRANGE0];
				
			if (YRANGE1 in data)
				y1 = data[YRANGE1];
		}
		
		return {
			init: function(graph) {
				graph.on('plotly_relayout', relayout);
			},
		
			reset: function() {
				x0 = left;
				x1 = right;
				y0 = bottom;
				y1 = top;
			},
			
			xrange: function() {
				return [x0, x1];
			},
			
			yrange: function() {
				return [y0, y1];
			},
			
			width: function() {
				return x1 - x0;
			},
			
			height: function() {
				return y1 - y0;
			},
			
			mulx: function(factor) {
				x0 *= factor;
				x1 *= factor;
			},
			
			muly: function(factor) {
				y0 *= factor;
				y1 *= factor;
			}
		};
	}

	return {
		MARGIN: { l: MARGIN_VERT, t: MARGIN_HORZ_EMPTY, r: MARGIN_VERT_EMPTY, b: MARGIN_HORZ },
		MARGIN_SECONDARY: { l: MARGIN_VERT, t: MARGIN_HORZ_EMPTY, r: MARGIN_VERT, b: MARGIN_HORZ },
		MARGIN_ZERO: { l: 0, t: 0, r: 0, b: 0 },
		MARGIN_NOTITLE: { l: MARGIN_HORZ, t: MARGIN_HORZ_EMPTY, r: MARGIN_HORZ_EMPTY, b: MARGIN_HORZ },
		MARGIN_CANVAS: { l: MARGIN_HORZ, t: 0, r: 0, b: MARGIN_HORZ },
		FONT: FONT,
		LABEL: { bgcolor: '#FFF', font: FONT },
		THIN_LINE: THIN_LINE,
		NORMAL_LINE: NORMAL_LINE,
		THICK_LINE: THICK_LINE,
	
		RangeManager: RangeManager,
	
		plot: function (x, y, config) {
			var line = {
				color: config && config.color ? config.color : 'red',
				simplify: false,
				width: config && config.width ? config.width : THICK_LINE};
		
			if (config && config.dash)
				line.dash = config.dash;
		
			var res = {
				x: x,
				y: y,
				name: config && config.name ? config.name : '',
				yaxis: config && (config.secondary || config.yaxis === 'y2') ? 'y2' : 'y',
				showlegend: config && config.legend !== undefined ? config.legend : true,
				line: line,
				mode: 'lines',
				type: 'scatter' };
				
			if (config && config.fill)
				res.fill = config.fill;
			else if (config && config.fillcolor)
				res.fill = 'tozeroy';

			if (config && config.fillcolor)
				res.fillcolor = config.fillcolor;

			return res;
		},
	
		solid: function (x, y, config) {
			return this.plot(x, y, config);
		},
	
		dashed: function (x, y, config) {
			if (!config)
				config = { dash: 'dash' };
			else
				config.dash = 'dash';
		
			return this.plot(x, y, config);
		},
	
		dotted: function (x, y, config) {
			if (!config)
				config = { dash: 'dot' };
			else
				config.dash = 'dot';
		
			return this.plot(x, y, config);
		},
	
		scatter: function (x, y, config) {
			return {
				x: x,
				y: y,
				name: config && config.name ? config.name : '',
				yaxis: config && config.secondary ? 'y2' : 'y',
				marker: {
					color: config && config.color ? config.color : 'red',
					size: config && config.size ? config.size : 10
					},
				showlegend: config && config.legend !== undefined ? config.legend : true,
				mode: 'markers',
				type: 'scatter' };
		},
	
		bar: function (x, y, config) {
			return {
				x: x,
				y: y,
				name: config && config.name ? config.name : '',
				type: 'bar',
				width: config && config.width ? config.width : 0.8,
				marker: {
					color: config && config.color ? config.color : 'red',
					opacity: config && config.opacity ? config.opacity : 0.5
				} };
		},
	
		filled: function (x, y, config) {
			return {
				x: x,
				y: y,
				name: config && config.name ? config.name : '',
				yaxis: config && config.secondary ? 'y2' : 'y',
				line: {
					color: config && config.color ? config.color : 'red',
					simplify: true,
					width: config && config.width ? config.width : THICK_LINE},
				mode: 'lines',
				type: 'scatter',
				fill: 'toself' };
		},
	
		HEATMAP: {
			type: 'heatmap',
			showscale: false,
			zsmooth: 'fast',
			hoverinfo: 'none'
		},
	
		heatmap: function(z, config) {
			if (!config)
				config = { };
			else if (config.colors) {
				config.colorscale = config.colors;
				delete config.colors;
			}
		
			config.z = z;
		
			return Object.fillEmpty(config, this.HEATMAP);
		},
	
		AXIS: {
		    	titlefont: FONT,
		    	tickfont: FONT,
		    	showgrid: true,
		    	ticks: 'outside',
		    	ticklen: 8,
		    	tickwidth: AXIS_LINE,
		    	zeroline: false,
		    	zerolinewidth: AXIS_LINE,
		    	showline: true,
		    	linewidth: AXIS_LINE,
		    	gridcolor: 'silver',
		    	gridwidth: THIN_LINE,
		    	exponentformat: 'power',
		    	hoverformat: '.4f' },
	
		axis: function (properties) {
			return Object.fillEmpty(properties, this.AXIS);
		},
	
		CONFIG: { scrollZoom: true, displayModeBar: false, responsive: true },
		
		config: function (properties) {
			return Object.fillEmpty(properties, this.CONFIG)
		},
	
		ARROW: {
			xref: 'x',
	 		yref: 'y',
	 		axref: 'x',
	 		ayref: 'y',
	 		font: FONT,
	 		showarrow: true,
			arrowhead: 2,
			arrowsize: 1.5,
			arrowwidth: THICK_LINE,
			arrowcolor: 'black',
			text: ''
		},
		
		arrow: function (properties) {
			return Object.fillEmpty(properties, this.ARROW)
		},
		
		harrow: function(x, y, ax, config) {
			var xref = config && config.xref ? config.xref : 'x';
			var yref = config && config.yref ? config.yref : 'y';
		
			return {
				x: x,
				y: y,
				ax: ax,
				ay: y,
				xref: xref,
		 		yref: yref,
		 		axref: xref,
		 		ayref: yref,
		 		font: FONT,
		 		showarrow: true,
				arrowhead: 2,
				arrowsize: 1.5,
				arrowwidth: config && config.width ? config.width : THICK_LINE,
				arrowcolor: config && config.color ? config.color : 'black',
				text: config && config.text ? config.text : ''
			};
		},
		
		varrow: function(x, y, ay, config) {
			var xref = config && config.xref ? config.xref : 'x';
			var yref = config && config.yref ? config.yref : 'y';
		
			return {
				x: x,
				y: y,
				ax: x,
				ay: ay,
				xref: xref,
		 		yref: yref,
		 		axref: xref,
		 		ayref: yref,
		 		font: FONT,
		 		showarrow: true,
				arrowhead: 2,
				arrowsize: 1.5,
				arrowwidth: config && config.width ? config.width : THICK_LINE,
				arrowcolor: config && config.color ? config.color : 'black',
				text: config && config.text ? config.text : ''
			};
		},

		vector: function(v, config) {
			return {
				x: v.x,
				y: v.y,
				ax: 0,
				ay: 0,
				xref: 'x',
		 		yref: 'y',
		 		axref: 'x',
		 		ayref: 'y',
		 		showarrow: true,
				arrowhead: 2,
				arrowsize: 1.5,
				arrowwidth: config && config.width ? config.width : THICK_LINE,
				arrowcolor: config && config.color ? config.color : 'black',
				text: config && config.text ? config.text : ''
			};
		},
		
		line: function (v1, v2, config) {
			return {
				type: 'line',
				xref: 'x',
				yref: 'y',
	        	x0: v1.x,
	        	y0: v1.y,
	        	x1: v2.x,
	        	y1: v2.y,
	        	line: {
			        color: config && config.color ? config.color : 'black',
			        width: config && config.width ? config.width : THICK_LINE,
			        dash: config && config.style ? config.style : 'dash'
			    }
	    	};
		},
		
		hline: function (y, config) {
			return {
				type: 'line',
				xref: 'paper',
				yref: 'y',
	        	x0: 0,
	        	y0: y,
	        	x1: 1,
	        	y1: y,
	        	line: {
			        color: config && config.color ? config.color : 'black',
			        width: config && config.width ? config.width : THICK_LINE,
			        dash: config && config.style ? config.style : 'dash'
			    }
	    	};
		},
		
		vline: function (x, config) {
			return {
				type: 'line',
				xref: config && config.xref ? config.xref : 'x',
				yref: config && config.yref ? config.yref : 'paper',
	        	x0: x,
	        	y0: config && config.y0 !== undefined ? config.y0 : 0,
	        	x1: x,
	        	y1: config && config.y1 !== undefined ? config.y1 : 1,
	        	line: {
			        color: config && config.color ? config.color : 'black',
			        width: config && config.width ? config.width : THICK_LINE,
			        dash: config && config.style ? config.style : 'dash'
			    }
	    	};
		},
		
		ANNOT: {
			xref: 'x', yref: 'x',
			xanchor: 'left', yanchor: 'top',
			font: FONT, showarrow: false },
		
		annot: function(x, y, text, config) {
			if (!config)
				config = { };
			else if (config.anchor) {
				var anchor = config.anchor;
				if (anchor.includes('r'))
					config.xanchor = 'right';
				else if (anchor.includes('c'))
					config.xanchor = 'center';
					
				if (anchor.includes('b'))
					config.yanchor = 'bottom';
				else if (anchor.includes('m'))
					config.yanchor = 'middle';
					
				delete config.anchor;
			}
		
			config.x = x;
			config.y = y;
			config.text = text;
		
			return Object.fillEmpty(config, this.ANNOT);
		},
		
		PTEXT: {
			xref: 'paper', yref: 'paper',
			xanchor: 'left', yanchor: 'top',
			font: FONT, showarrow: false },
		
		ptext: function(x, y, text, config) {
			if (!config)
				config = { };
			else if (config.anchor) {
				var anchor = config.anchor;
				if (anchor.includes('r'))
					config.xanchor = 'right';
				else if (anchor.includes('c'))
					config.xanchor = 'center';
					
				if (anchor.includes('b'))
					config.yanchor = 'bottom';
				else if (anchor.includes('m'))
					config.yanchor = 'middle';
					
				delete config.anchor;
			}
		
			config.x = x;
			config.y = y;
			config.text = text;
		
			return Object.fillEmpty(config, this.PTEXT);
		},
		
		vertImage: function(source, x1, x2, opacity) {
			return {
		        source: source,
		        xref: 'x',
		        yref: 'paper',
		        x: x1,
		        y: 0,
		        sizex: x2 - x1,
		        sizey: 1,
		        xanchor: 'left',
		        yanchor: 'bottom',
		        sizing: 'stretch',
		        layer: 'below',
		        opacity: opacity ? opacity : 0.5
		    };
		},
		
		dtick: function(range, factor) {
			var n = Math.floor(Math.log10(range));
			var om = Math.pow(10, n);
			range /= om;
			var delta = 0.2;
			if (range > factor * 2)
				delta = 2
			else if (range > factor)
				delta = 1;
			else if (range > factor * 0.5)
				delta = 0.5;
			
			return delta * om;
		},
		
		save: function(graph, filename) {
			Plotly.downloadImage(graph, {format: 'svg', filename: filename});
		},
		
		autorange: function(graph) {
			Plotly.relayout(graph, { 'xaxis.autorange': true, 'yaxis.autorange': true });
		},
		
		grid: function(graph, show) {
			Plotly.relayout(graph, { 'xaxis.showgrid': show, 'yaxis.showgrid': show });
		}
	};
})();