﻿"use strict";

function Path(p) {
	this.x = [p.x];
	this.y = [p.y];
};

Path.prototype.push = function(p) {
	this.x.push(p.x);
	this.y.push(p.y);
};

var TRACE = {
	MIN_LENGTH: 1e-10,
	EPS: 1e-12,
	MAX_DIST: 200,
	MAX_IRC: 10,   // maximum internal reflection count
	
	trace: function(lightRay, elements) {
		var ray, wave;
		if (lightRay.ray) {
			ray = lightRay.ray;
			wave = lightRay.wave;
		} else {
			ray = lightRay;
			wave = undefined;
		}
			
		var path = new Path(ray.pos);
	
		/*elements.forEach(function(e){
			ray = e.trace(ray, wave, path);
			if (!ray)
				return path;
		});*/
	
		for (var i = 0; i < elements.length; i++) {
			ray = elements[i].trace(ray, wave, path);
			if (!ray)
				return path;
		}
	
		var p = ray.move(TRACE.MAX_DIST);
		path.push(p);
		return path;
	}
};


function Ray(pos, dir) {
	this.pos = pos;
	this.dir = dir;
};

Ray.prototype.rotate = function(angle, origin) {
	var pos = this.pos;
	if (origin)
		pos = pos.sub(origin)
	pos = pos.rotate(angle);
	if (origin)
		pos = pos.add(origin);
	
	return new Ray(pos, this.dir.rotate(angle));
};
	
Ray.prototype.move = function(dist) {
	return this.pos.add(this.dir.unit().mul(dist));
};


var LightRay = (function(){
	var NODES_LIGHT = [
		{wave: 0.35, color: [255, 0, 255]},
		{wave: 0.45, color: [51, 51, 255]},
		{wave: 0.5,  color: [0, 255, 255]},
		{wave: 0.55, color: [0, 255, 0]},
		{wave: 0.6,  color: [255, 255, 0]},
		{wave: 0.65, color: [255, 0, 0]},
		{wave: 0.75, color: [255, 0, 128]}
	];

	var NODES_DARK = [
		{wave: 0.35, color: [102, 0, 204]},
		{wave: 0.45, color: [0, 0, 255]},
		{wave: 0.5,  color: [0, 204, 204]},
		{wave: 0.55, color: [0, 204, 0]},
		{wave: 0.6,  color: [204, 204, 0]},
		{wave: 0.65, color: [255, 0, 0]},
		{wave: 0.75, color: [204, 0, 102]}
	];

	var NODES = NODES_LIGHT;
	
	function rgbToHtml(rgb) {
		return '#' + rgb[0].toString(16).padStart(2,'0') + rgb[1].toString(16).padStart(2,'0') + rgb[2].toString(16).padStart(2,'0');
	}

	function waveToHtml(wave) {
		if (wave <= NODES[0].wave)
			return rgbToHtml(NODES[0].color);
		
		var last = NODES.length - 1;
		for (var i = 0; i < last; i++)
			if (wave <= NODES[i + 1].wave) {
				var rgb = [0, 0, 0];
				var fraction = (wave - NODES[i].wave) / (NODES[i + 1].wave - NODES[i].wave);
				for (var k = 0; k < 3; k++)
					rgb[k] = Math.round(NODES[i].color[k] + fraction * (NODES[i + 1].color[k] - NODES[i].color[k]));
				return rgbToHtml(rgb);
			}
			
		return rgbToHtml(NODES[last].color);
	}

	function constructor(ray, wave) {
		this.ray = ray;
		this.wave = wave;	
	};

	constructor.darkColors = function() {
		NODES = NODES_DARK;
	};

	constructor.lightColors = function() {
		NODES = NODES_LIGHT;
	};

	constructor.prototype.color = function() {
		return waveToHtml(this.wave);
	}

	return constructor;
})();


function Segment(n1, n2) {
	this.n1 = n1;
	this.n2 = n2;
};

Segment.prototype.direction = function() {
	return this.n2.sub(this.n1);
};
		
Segment.prototype.length = function() {
	return this.direction().length();
};

Segment.prototype.center = function() {
	return this.n1.add(this.n2).mul(0.5);
};

Segment.prototype.intersect = function(ray) {
	var p_r = ray.pos;
	var u_r = ray.dir.unit();
	var p_c = this.center();
	var v_t = this.direction();
	var u_n = v_t.perp().unit();
	
	var dot = u_r.dot(u_n);
	if (Math.abs(dot) < TRACE.EPS) // praktiliselt paralleelne pinnaga
		return null;
	
	var dist = p_c.sub(p_r).dot(u_n) / dot;
	if (dist < TRACE.MIN_LENGTH) // isegi kui on natuke positiivne
		return null;
	
	var p_i = p_r.add(u_r.mul(dist));
	if (p_c.sub(p_i).length() > 0.5 * v_t.length() + TRACE.MIN_LENGTH)
		return null;
	
	return {point: p_i, dist: dist, norm: u_n};
};


function Polyline(x, y) {
	this.x = x;
	this.y = y;
}

Polyline.prototype.getSegment = function(index) {
	return new Segment(
		new Vector(this.x[index], this.y[index]),
		new Vector(this.x[index + 1], this.y[index + 1])
	);
};

Polyline.prototype.getSegments = function() {
	var res = [];
	for (var i = 0; i < this.x.length - 1; i++)
		res.push(this.getSegment(i));
	return res;
};

Polyline.prototype.intersect = function(ray) {
	var closest = null;
	
	var segments = this.getSegments();
	for (var i = 0; i < segments.length; i++) {
		var ir = segments[i].intersect(ray);
		if (ir && (closest == null || closest.dist > ir.dist)) {
			closest = ir;
			closest.segment = segments[i];
		}
	}
	
	return closest;
};

Polyline.prototype.translate = function(dx, dy) {
	var nx = [];
	var ny = [];
	
	for (var i = 0; i < this.x.length; i++) {
		nx.push(this.x[i] + dx);
		ny.push(this.y[i] + dy);
	}
	
	return new Polyline(nx, ny);
};

Polyline.prototype.rotate = function(angle) {
	var sin = Math.sin(angle);
	var cos = Math.cos(angle);
	var nx = [];
	var ny = [];
	
	for (var i = 0; i < this.x.length; i++) {
		nx.push(this.x[i] * cos - this.y[i] * sin);
		ny.push(this.x[i] * sin + this.y[i] * cos);
	}
	
	return new Polyline(nx, ny);
};

Polyline.createPrism = function(height, angle) {
	var a = height * Math.tan(0.5 * angle);
	return new Polyline([-a, 0, a, -a], [0, height, 0, 0]);
};

Polyline.createLittrow = function(height, angle) {
	var width = height / Math.tan(angle);
	return new Polyline([0, width, 0, 0], [0, 0, height, 0]);
};

Polyline.createRhomb = function(width, height, angle) {
	var a = 0.5 * width;
	var b = 0.5 * height;
	var d = angle ? b * Math.tan(angle) : 0;
	return new Polyline([-a + d, a + d, a - d, -a - d, -a + d], [b, b, -b, -b, b]);
};

Polyline.createLens = function(diam, angle) {
	angle = 0.5 * angle;
	var offset = 0.5 * diam / Math.tan(angle);
	var radius = 0.5 * diam / Math.sin(angle);
	//var angle = Math.asin(0.5 * diam / radius);
	var delta = 0.5 * Math.DEG;
	var count = Math.ceil(2 * angle / delta);
	delta = 2 * angle / count;
	var x = [0];
	var y = [-0.5 * diam];
	
	function makeArc(center, start) {
		for (var i = 1; i <= count; i++) {
			var phi = start + i * delta;
			x.push(center + radius * Math.cos(phi));
			y.push(radius * Math.sin(phi));				
		}
	};

	makeArc(-offset, -angle);
	makeArc(offset, Math.PI - angle);
	return new Polyline(x, y);
};


function Straight(n1, n2) {
	this.n1 = n1;
	this.n2 = n2;
};

Straight.prototype.collect = function(x, y) {
	x.push(this.n1.x);
	y.push(this.n1.y);
	x.push(this.n2.x);
	y.push(this.n2.y);
};

Straight.prototype.direction = function() {
	return this.n2.sub(this.n1);
};

Straight.prototype.length = function() {
	return this.direction().length();
};

Straight.prototype.center = function() {
	return this.n1.add(this.n2).mul(0.5);
};

Straight.prototype.translate = function(dx, dy) {
	this.n1 = this.n1.translate(dx, dy);
	this.n2 = this.n2.translate(dx, dy);
	return this;
};

Straight.prototype.rotate = function(angle) {
	this.n1 = this.n1.rotate(angle);
	this.n2 = this.n2.rotate(angle);
	return this;
};

Straight.prototype.intersect = function(ray) {
	var p_r = ray.pos;
	var u_r = ray.dir.unit();
	var p_c = this.center();
	var v_t = this.direction();
	var u_n = v_t.perp().unit();
	
	var d = u_r.dot(u_n);
	if (Math.abs(d) < TRACE.EPS) // praktiliselt paralleelne pinnaga
		return null;
	
	var dist = p_c.sub(p_r).dot(u_n) / d;
	if (dist < TRACE.MIN_LENGTH) // isegi kui on natuke positiivne
		return null;
	
	var p_i = p_r.add(u_r.mul(dist));
	if (p_c.sub(p_i).length() > 0.5 * v_t.length() + TRACE.MIN_LENGTH)
		return null;
	
	return {point: p_i, dist: dist, norm: u_n};
};


function Curved(center, radius, start, alpha) {
	this.center = center;
	this.radius = radius;
	this.start = start;
	this.alpha = alpha; // võib olla ka negatiivne
};

Curved.prototype.collect = function(x, y) {
	var delta = 0.5 * Math.DEG;
	var count = Math.ceil(Math.abs(this.alpha) / delta);
	delta = this.alpha / count;
	for (var i = 0; i <= count; i++) {
		var phi = this.start + i * delta;
		x.push(this.center.x + this.radius * Math.cos(phi));
		y.push(this.center.y + this.radius * Math.sin(phi));				
	}
};

Curved.prototype.translate = function(dx, dy) {
	this.center = this.center.translate(dx, dy);
	return this;
};

Curved.prototype.rotate = function(angle) {
	this.center = this.center.rotate(angle);
	this.start += angle;
	return this;
};

Curved.prototype.intersect = function(ray) {
	var p_c = this.center;
	var alpha = this.alpha;
	var start = this.start;
	var p_r = ray.pos;
	var u_r = ray.dir.unit();
	var u_p = u_r.perp();
	var t_c = p_c.sub(p_r);
	var d_p = Math.abs(t_c.dot(u_p));
	if (d_p > this.radius)
		return null;
		
	var delta = Math.sqrt(this.radius * this.radius - d_p * d_p);
	var d_r = t_c.dot(u_r);
	
	function check(dist) {
		if (dist < TRACE.MIN_LENGTH) // isegi kui on positiivne (?)
			return null;
			
		var p_i = p_r.add(u_r.mul(dist));
		var v_n = p_i.sub(p_c);
		var angle = (v_n.angle() - start) * Math.sign(alpha);
		angle -= Math.floor(angle / Math.DPI) * Math.DPI;
			if (angle > Math.abs(alpha))
				return null;
			
		return {point: p_i, dist: dist, norm: v_n};
	};
	
	var ir = check(d_r - delta);
	if (!ir)
		ir = check(d_r + delta);
	if (!ir)
		return null;
	
	return ir;
};


function Element(sections) {
	this.sections = sections;
};

Element.prototype.draw = function() {
	var x = [];
	var y = [];
	for (var i = 0; i < this.sections.length; i++)
		this.sections[i].collect(x, y);
		
	return new Polyline(x, y);
};

Element.prototype.intersect = function(ray) {
	var closest = null;
	
	for (var i = 0; i < this.sections.length; i++) {
		var section = this.sections[i];
		var ir = section.intersect(ray);
		if (ir && (closest == null || closest.dist > ir.dist)) {
			closest = ir;
			closest.section = section;
		}
	}
	
	return closest;
};

Element.prototype.translate = function(dx, dy) {
	for (var i = 0; i < this.sections.length; i++)
		this.sections[i].translate(dx, dy);
		
	return this;
};

Element.prototype.rotate = function(angle) {
	for (var i = 0; i < this.sections.length; i++)
		this.sections[i].rotate(angle);
		
	return this;
};

Element.prototype.trace = function(ray, wave, path) {
	var p, ir;
	
	ir = this.intersect(ray);
	if (!ir)
		return ray;
	
	var ridx = wave ? this.dispers(wave) : this.ridx;
	p = ir.point;
	path.push(p);
	ir = ray.dir.refract(ir.norm, ridx);
	ray = new Ray(p, ir.dir);
	if (!ir.trans)
		return ray;
	
	var count = 0;
	while (true) {
		ir = this.intersect(ray);
		if (!ir)
			return null;
		
		p = ir.point;
		path.push(p);
		
		ir = ray.dir.refract(ir.norm, 1 / ridx);
		ray = new Ray(p, ir.dir);
		
		if (ir.trans)
			return ray;
		
		count++;

		if (count > TRACE.MAX_IRC)
			return null;
	}
};

function IsoscelesPrism(height, angle) {
	var a = height * Math.tan(0.5 * angle);
	
	var n1 = new Vector(-a, 0);
	var n2 = new Vector(a, 0);
	var n3 = new Vector(0, height);
	
	return new Element([
		new Straight(n1, n2),
		new Straight(n2, n3),
		new Straight(n3, n1)
	]);
}

function Littrow(height, angle) {
	var width = height * Math.tan(angle);
	var n1 = new Vector(0, 0);
	var n2 = new Vector(width, 0);
	var n3 = new Vector(0, height);
	
	return new Element([
		new Straight(n1, n2),
		new Straight(n2, n3),
		new Straight(n3, n1)
	]);
}

function Rhomb(width, height, angle) {
	var a = 0.5 * width;
	var b = 0.5 * height;
	var d = angle ? b * Math.tan(angle) : 0;
	
	var n1 = new Vector(-a - d, -b);
	var n2 = new Vector(a - d, -b);
	var n3 = new Vector(a + d, b);
	var n4 = new Vector(-a + d, b);
	
	return new Element([
		new Straight(n1, n2),
		new Straight(n2, n3),
		new Straight(n3, n4),
		new Straight(n4, n1)
	]);
}

function SpheriCube(ir, width, length) {
	width *= 0.5;
	var n1 = new Vector(0, -width);
	var n2 = new Vector(length, -width);
	var n3 = new Vector(length, width);
	var n4 = new Vector(0, width);

	var surfaces = [
		new Straight(n1, n2),
		new Straight(n2, n3),
		new Straight(n3, n4)
	];
	
	if (Math.abs(ir) * width < TRACE.EPS)
	{
		surfaces.push(new Straight(n4, n1));
		var lens = new Element(surfaces);
		
		lens.cardinal = function(index) {
			return {f: index / (ir * (index - 1)), vertex: 0, center: 1 / ir};
		};
		
		return lens;
	}
	
	var radius = 1 / Math.abs(ir);
	if (Math.abs(radius - width) < TRACE.MIN_LENGTH)
	{
		surfaces.push(new Curved(new Vector(0, 0), radius, Math.HPI, Math.sign(ir) * Math.PI));
		var lens = new Element(surfaces);
		
		lens.cardinal = function(index) {
			return {f: index / (ir * (index - 1)), vertex: -Math.sign(ir) * radius, center: 0};
		};
		
		return lens;
	}

	if (radius < width)
	{
		surfaces.push(new Straight(n4, new Vector(0, radius)));
		surfaces.push(new Curved(new Vector(0, 0), radius, Math.HPI, Math.sign(ir) * Math.PI));
		surfaces.push(new Straight(new Vector(0, -radius), n1));
		
		var lens = new Element(surfaces);
		
		lens.cardinal = function(index) {
			return {f: index / (ir * (index - 1)), vertex: -Math.sign(ir) * radius, center: 0};
		};
		
		return lens;
	}
	
	var offset = Math.sign(ir) * Math.sqrt(radius * radius - width * width);
	var center = new Vector(offset, 0);
	var half = Math.asin(width / radius);
	var full = 2 * half;
	if (ir > 0)
		surfaces.push(new Curved(center, radius, Math.PI - half, full));
	else
		surfaces.push(new Curved(center, radius, half, -full));
	
	var lens = new Element(surfaces);
		
	lens.cardinal = function(index) {
		return {f: index / (ir * (index - 1)), vertex: offset - Math.sign(ir) * radius, center: offset};
	};
	
	return lens;	
}

function GenericLens(diam, ir1, ir2, thickn) {
	var hd = 0.5 * diam;
	var ht = 0.5 * thickn;
	var hds = hd * hd;
	
	function f(x) {
		return thickn - ir1 * x / (1 + Math.sqrt(1 - ir1 * ir1 * x)) - ir2 * x / (1 + Math.sqrt(1 - ir2 * ir2 * x));
	}
	
	var sidelen = f(hds);
	if (sidelen < 0) {
		hds = Math.secant(f, hds, 0.5 * hds, 1e-10);
		hd = Math.sqrt(hds);
		//console.log( hd );
	}
	
	var s1, s2, half, center, radius, lc, rc;
	
	half = Math.asin(hd * ir1);
	if (Math.abs(half) < TRACE.EPS) {
		lc = -ht;
		s1 = new Straight(new Vector(lc, hd), new Vector(lc, -hd));
	} else {
		radius = 1 / ir1;
		center = radius - ht;
		lc = center - radius * Math.cos(half);
		s1 = new Curved(new Vector(center, 0), Math.abs(radius), half > 0 ? Math.PI - half : -half, 2 * half);
	}
	
	half = Math.asin(hd * ir2);
	if (Math.abs(half) < TRACE.EPS) {
		rc = ht;
		s2 = new Straight(new Vector(rc, -hd), new Vector(rc, hd));
	} else {
		radius = 1 / ir2;
		center = -radius + ht;
		rc = center + radius * Math.cos(half);
		s2 = new Curved(new Vector(center, 0), Math.abs(radius), half > 0 ? -half : Math.PI - half, 2 * half);
	}
	
	var lens = [s1];
	if (sidelen > TRACE.MIN_LENGTH)
		lens.push( new Straight(new Vector(lc, -hd), new Vector(rc, -hd)) );
	lens.push( s2 );
	if (sidelen > TRACE.MIN_LENGTH)
		lens.push( new Straight(new Vector(rc, hd), new Vector(lc, hd)) );
	
	var lens = new Element(lens);
	
	lens.cardinal = function(index) {
		var tmp = (index - 1) * thickn / index;
		var focal = 1 / ( (index - 1) * (ir1 + ir2 - tmp * ir1 * ir2) );
		tmp *= focal;
		return {f: focal, h1: -ht + tmp * ir2, h2: ht - tmp * ir1};
	};
		
	return lens;
}

function DoubleConvex(diam, thickn) {
	var center = (diam * diam - thickn * thickn) / (4 * thickn);
	var radius = center + 0.5 * thickn;
	var half = Math.asin(0.5 * diam / radius);
	var full = 2 * half;
	
	var lens = new Element([
		new Curved(new Vector(-center, 0), radius, -half, full),
		new Curved(new Vector(center, 0), radius, Math.PI - half, full)
	]);
	
	lens.cardinal = function(index) {
		var tmp = (index - 1) * thickn / (index * radius);
		var focal = 1 / ( (index - 1) * (2 - tmp) / radius );
		var offset = focal * tmp;
		return {f: focal, h1: -0.5 * thickn + offset, h2 : 0.5 * thickn - offset};
	};
		
	return lens;
}

function DoubleConvex2(diam, angle) {
	var half = 0.5 * angle;
	var offset = 0.5 * diam / Math.tan(half);
	var radius = 0.5 * diam / Math.sin(half);
	
	return new Element([
		new Curved(new Vector(-offset, 0), radius, -half, angle),
		new Curved(new Vector(offset, 0), radius, Math.PI - half, angle)
	]);
}

function PlanoConvex(diam, thickn) {
	var radius = (diam * diam - 4 * thickn * thickn) / (8 * thickn) + thickn;
	var half = Math.asin(0.5 * diam / radius);
	var full = 2 * half;
	
	var lens = new Element([
		new Curved(new Vector(-radius, 0), radius, -half, full),
		new Straight(new Vector(-thickn, 0.5 * diam), new Vector(-thickn, -0.5 * diam))
	]);
	
	lens.cardinal = function(index) {
		var focal = radius / (index - 1);
		var offset = focal * (index - 1) * thickn / (index * radius);
		return {f: focal, h1: offset - thickn, h2 : 0};
	};
		
	return lens;
}

function RevPlanoConvex(diam, thickn) {
	var radius = (diam * diam - 4 * thickn * thickn) / (8 * thickn) + thickn;
	var half = Math.asin(0.5 * diam / radius);
	var full = 2 * half;
	
	var lens = new Element([
		new Curved(new Vector(radius, 0), radius, Math.PI - half, full),
		new Straight(new Vector(thickn, -0.5 * diam), new Vector(thickn, 0.5 * diam))
	]);
	
	lens.cardinal = function(index) {
		var focal = radius / (index - 1);
		var offset = focal * (index - 1) * thickn / (index * radius);
		return {f: focal, h1: 0, h2 : -offset + thickn};
	};
		
	return lens;
}

function BallLens(diam) {
	var radius = 0.5 * diam;
	var lens = new Element([ new Curved(new Vector(0, 0), radius, 0, Math.DPI) ]);

	lens.cardinal = function(index) {
		var tmp = 2 * (index - 1) / index;
		var focal = 1 / ( (index - 1) * (2 - tmp) / radius );
		var offset = focal * tmp;
		return {f: focal, h1: -radius + offset, h2 : radius - offset};
	};
		
	return lens;
}

function Shell(diam, radius, thickn) {
	var half = Math.asin(0.5 * diam / radius);
	var full = 2 * half;
	
	return new Element([
		new Curved(new Vector(0, 0), radius, -half, full),
		new Straight(Vector.polar(radius, half), Vector.polar(radius + thickn, half)),
		new Curved(new Vector(0, 0), radius + thickn, half, -full),
		new Straight(Vector.polar(radius + thickn, -half), Vector.polar(radius, -half))
	]);	
}

