﻿"use strict";

var Taylor = (function() {
	function Range(min, max) {
		this.min = min;
		this.max = max;
	}
	
	Range.prototype.span = function() {
		return this.max - this.min;
	};
	
	Range.prototype.extend = function(factor) {
		var span = this.span();
		var delta = 0.5 * span * (factor - 1);
		return new Range(this.min - delta, this.max + delta);
	};

	Range.prototype.intersect = function(range) {
		if (range.min > this.max || range.max < this.min)
			return null;
		
		return new Range(Math.max(this.min, range.min), Math.min(this.max, range.max));
	};

	// suitable for sin, cos
	function defaultView() {
		return {
			x: new Range(-6, 6),
			y: new Range(-2, 2)
		};
	}
	
	function T(data) {
		var valid = 'valid' in data ? data.valid : function(x) { return true; };
	
		this.valid = valid;
	
		this.eval = function(x) { return valid(x) ? data.eval(x) : NaN; };
	
		this.getComp = function(n) {
			var order = data.order(n);
			var factor = data.deriv(n) / Math.fact(order);
			
			return function(x) {
				return factor * Math.pow(x, order);
			};
		};
		
		this.getPartial = function(order) {
			var comp = [];
			for (var n = 0; n <= order; n++)
				comp.push( this.getComp(n) );
		
			return function(x) {
				var result = 0;
				for (var n = 0; n <= order; n++)
					result += comp[n](x);
				return result;
			};
		};
		
		this.getOrder = data.order;
		this.getDeriv = data.deriv;
		this.getView = function() { return 'view' in data ? data.view : defaultView(); };
		this.getMarks = function() { return 'marks' in data ? data.marks : []; };
	}

	T.Sin = function() {
		return new T({
			eval: function(x) { return Math.sin(x); },
			order: function(n) { return n ? 2 * n - 1 : 0; },
			deriv: function(n) { return n ? Math.pow(-1, n + 1) : 0; }
		});
	};
	
	T.Cos = function() {
		return new T({
			eval: function(x) { return Math.cos(x); },
			order: function(n) { return 2 * n; },
			deriv: function(n) { return Math.pow(-1, n); }
		});
	};

	T.Sinh = function() {
		return new T({
			eval: function(x) { return Math.sinh(x); },
			order: function(n) { return n ? 2 * n - 1 : 0; },
			deriv: function(n) { return n ? 1 : 0; },
			view: { x: new Range(-4, 4), y: new Range(-20, 20) }
		});
	};

	T.Exp = function() {
		return new T({
			eval: function(x) { return Math.exp(x); },
			order: function(n) { return n; },
			deriv: function(n) { return 1; },
			view: { x: new Range(-1.5, 1.5), y: new Range(0, 5.5) }
		});
	};
	
	T.Ln = function() {
		return new T({
			eval: function(x) { return Math.log(1 + x); },
			order: function(n) { return n; },
			deriv: function(n) { return n ? Math.pow(-1, n - 1) * Math.fact(n - 1) : 0; },
			valid: function(x) { return x > -0.99; },
			view: { x: new Range(-1, 1.5), y: new Range(-3, 1) },
			marks: [-1, 1]
		});
	};
	
	T.Geom = function() {
		return new T({
			eval: function(x) { return 1 / (1 - x); },
			order: function(n) { return n; },
			deriv: function(n) { return Math.fact(n); },
			valid: function(x) { return x < 0.99 || x > 1.01; },
			view: { x: new Range(-1, 1), y: new Range(0, 5) },
			marks: [-1, 1]
		});
	};

	return T;
})();